import gcprofiler
import json
import gc

class Node:
    def __init__(self, value):
        self.value = value
        self.next = None

class Person:
    def __init__(self, name):
        self.name = name
        self.friends = []

def leakMemory():
    # Create circular references
    node1 = Node(1)
    node2 = Node(2)
    node3 = Node(3)
    node1.next = node2
    node2.next = node3
    node3.next = node1  # Circular reference
    person1 = Person("Alice")
    person2 = Person("Bob")
    person1.friends.append(person2)
    person2.friends.append(person1)





    # Force garbage collection
    gc.collect()

    # Create a dictionary to store the leaked objects
    leaked_objects = []

    # Check for objects in gc.garbage
    if gc.garbage:
       for obj in gc.garbage:
           leaked_objects.append(repr(obj))

    # Store the leaked objects in a dictionary
    result = {"leaked_objects": leaked_objects}

    # Convert the dictionary to JSON format
    json_result = json.dumps(result, indent=4)

# Print the JSON result
    print(json_result)





import subprocess
import os
import pexpect
import time
from log.ycrash_logger import ycrash_logger

"""
Execute a system command and return its output and return code.

Args:
- command: str or list. The command to be executed. If it's a string, it will be passed to the shell for execution.

Returns:
- output: str. The output of the command.
"""
def run_command(command):
    try:
        if isinstance(command, str):
            output = subprocess.check_output(command, shell=True, universal_newlines=True)
        elif isinstance(command, list):
            output = subprocess.check_output(command, universal_newlines=True)
        else:
            raise ValueError("Invalid command type. Must be a string or list.")

        return_code = 0  # Success
    except subprocess.CalledProcessError as e:
        ycrash_logger.error(f" {command} failed with error {e.output} return_code {e.returncode}")
        output = f"Error while retriving for command {command} errorcode {e.returncode}"
        return_code = e.returncode
    return output.strip()

"""
    Get the current process ID (PID).

    Returns:
    - pid: int. The current process ID.
"""
def get_current_pid():
    return os.getpid()

"""
    Runs interactive commands, waits for the configured sleep time for the command to complete, and sends a carriage return to retrieve the system logs.
    Returns:
    - output: string. Runs the command output
"""
def run_interactive_command(command,sleeptime):
    # Run the 'top' command with the specified PID
    process = pexpect.spawn(command)
    time.sleep(sleeptime)
    process.send('\r')
    print(f"Carriaege retuurn")
    process.sendintr()
    output = ""
    try:
        # Capture output
        # output = process.read(size=5000).decode('utf-8')
        while True:
            chunk = process.read(100)
            output += chunk.decode('utf-8')
            if not chunk:
                break  # Reached EOF
    except pexpect.exceptions.TIMEOUT as e:
        print("Timeout exceeded while executing command:", command)
        return output
    # while True:
    # chunk = process.read(5000)
    #  if not chunk:
    #      break  # Reached EOF
    #   output += chunk.decode('utf-8')
    # print(f"read")

    # Return the captured output
    return output


"""
    Get the server API key from yCrash config

    Returns:
    - string. key from config file
"""
def getKey(config):
    return config.get('options', {}).get('k')


"""
    Get the server url  from yCrash config

    Returns:
    - string. url from config file
"""
def getServerUrl(config, endpoint):
    return config.get('options', {}).get('s') + endpoint



"""
    yCrash Python Profiler: Log Analyser
    This service responsible to extract the log changes and uploads to sever in chunks
"""

import os
import logger
import json
import time
from commons.webclient import upload_json_data
from commons.common import  getKey, getServerUrl
from .ycrash_logger import ycrash_logger


LOG_DUMP_ENDPOINT = "/cm-receiver+&dt=applog"
LOG_UPLOAD_LINES=100

"""
 Uploads the log changes to the server in chunks:
 - Log files are discovered or provided as arguments.
 - The log analyzer identifies the differences between backed-up and current log files.
"""
def upload_log_data(config):
    if config.get('options', {}).get('appLogs') is not None:
        log_path = config.get('options', {}).get('appLogs')
    else:
        log_path = get_log_file_location()

    logBackUpFile = log_path + '.ycrash.bak'
    if not os.path.exists(logBackUpFile):
        with open(logBackUpFile, 'w'):
            pass

    with open(log_path, 'r') as new_log_file:
        new_log_lines = set(new_log_file.readlines())

    with open(logBackUpFile, 'r') as old_log_file:
        old_log_lines = set(old_log_file.readlines())

    new_additions = new_log_lines - old_log_lines
    new_lines_list = list(new_additions)
    uploadInChunks(config, new_lines_list)
    copy_file(log_path, logBackUpFile)

"""
    Uploads the log  to the yCrash server in chunks
"""
def uploadInChunks(config, new_additions):
    for i in range(0, len(new_additions), LOG_UPLOAD_LINES):
        chunk_lines = new_additions[i:i + LOG_UPLOAD_LINES]
        log_lines = json.dumps(chunk_lines, indent=4)
        log_json = {
            "captureTime": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
            "logs": log_lines
        }

        # Upload log_json
        upload_json_data('application/json', getKey(config), getServerUrl(config, LOG_DUMP_ENDPOINT),
                         json.dumps(log_json, indent=4))

        # Make a backup of the log file after uploading each chunk


"""
    Retrieves log file location from logger
    Returns:
    -  str. The absolute path of the 
"""
def get_log_file_location():
    # Get the root logger
    root_logger = logger.getLogger()

    # Iterate over all handlers of the root logger
    for handler in root_logger.handlers:
        # Check if the handler is a FileHandler
        if isinstance(handler, logger.FileHandler):
            # Return the log file's absolute path
            return handler.baseFilename

    # If no FileHandler is found, return None or any other value you prefer
    return None

"""
    Copies aall the source to the destinations
    - used to backup log file
"""
def copy_file(source_file, destination_file):
    try:
        # Open the source file in binary mode for reading
        with open(source_file, 'rb') as src_file:
            # Open the destination file in binary mode for writing
            with open(destination_file, 'wb') as dest_file:
                # Read the content of the source file and write it to the destination file
                dest_file.write(src_file.read())
        ycrash_logger.debug("yCrash log backup file copied successfully.")
    except FileNotFoundError:
        ycrash_logger.error("File not found.")
    except Exception as e:
        ycrash_logger.error(f"An error occurred: {e}")

"""
    Retrieves all the loggers to the logging module
    Returns:
    -  list of str. The absolute path of all the attached logger to the python module.
"""
def get_all_loggers():
    root_logger = ycrash_logger.getLogger()
    all_loggers = [root_logger]

    for logger_name in ycrash_logger.Logger.manager.loggerDict.keys():
        logger = logger.getLogger(logger_name)
        if logger.parent is root_logger:
            all_loggers.append(logger)

    return all_loggers



"""
    yCrash Python Profiler: Profiler
    Main profiler service reads yCrash config, collected data and uploads to yCrash server
"""

import time
import threading
from threads.thread_profiler import capture_export_thread_data
from heap.heap_dump_profiler import HeapDumpCapturer
from process.process_profiler import capture_print_process_details
from gcprofiler.gc_profiler import ycrash_memory_extract
from log.log_profiler import upload_log_data
from system.system_profiler import export_all_system_data
from commons.webclient import upload_json_data
from commons.common import  getKey, getServerUrl
import yaml
import json
from log.ycrash_logger import ycrash_logger
from datetime import datetime


YC_FIN_ENDPOINT = "/cm-fin"


"""
    yCrash init - Routine to be called from external programs
    Runs a background thread to upload data.
"""

def ycrash_init(configFilePath):
    thread_states_thread = threading.Thread(target=profile_data, args=(configFilePath,), name="yCrash.analyzer")
    thread_states_thread.start()

"""
    Profiler reads the config and calls data collection and uploads
"""
def profile_data(configFilePath):

    # Calculate the time taken

    ycrashConfig = load_config(configFilePath)
    # gc.set_debug(gc.DEBUG_LEAK)
    # while True:
    start_time = datetime.now()
    upload_log_data(ycrashConfig)
    export_all_system_data(ycrashConfig)
    capture_export_thread_data(ycrashConfig)
    capturer = HeapDumpCapturer()
    capturer.capture_heap_dump(ycrashConfig)
    capture_print_process_details(ycrashConfig)
    ycrash_memory_extract(ycrashConfig)
    submit_fin(start_time, ycrashConfig)

        # time.sleep(ycrashConfig.get('options', {}).get('m3Frequency'))

"""
    Submits a fin after data is uploaded to the server
"""
def submit_fin(start_time, ycrashConfig):
    end_time = datetime.now()
    time_took = end_time - start_time
    time_taken = {
        "start_time": start_time.strftime("%Y-%m-%d %H:%M:%S"),
        "end_time": end_time.strftime("%Y-%m-%d %H:%M:%S"),
        "time_taken(seconds)": time_took.total_seconds()
    }
    time_json = json.dumps(time_taken, indent=4)
    upload_json_data('application/json', getKey(ycrashConfig), getServerUrl(ycrashConfig, YC_FIN_ENDPOINT), time_json)
    return time

"""
    Loads and parses yCrash config file 
"""
def load_config(configFilePath):
    try:
        # Read the YAML file
        with open(configFilePath, 'r') as file:
            print(file)
            data = yaml.safe_load(file)
        # Accessing attributes
        version = data.get('version')
        options = data.get('options', {})

        # Accessing specific attributes within options
        k = options.get('k')
        s = options.get('s')
        a = options.get('a')
        m3Frequency = options.get('m3Frequency')
        app_logs = options.get('appLogs', [])

        # Outputting the values
        ycrash_logger.debug(f"Version:{version}")
        # ycrash_logger.debug("key:", k)
        ycrash_logger.debug(f"Server(s): {s}")
        ycrash_logger.debug(f"Application(a): {a}")
        ycrash_logger.debug(f"Monitoing Frequency(m3Frequency): {m3Frequency}")

    except FileNotFoundError:
        ycrash_logger.error(f"yCrash agent config {configFilePath} not found")
        raise ValueError("yCrash agent config {configFilePath} not found")

    except yaml.YAMLError as exc:
        ycrash_logger.error("yCrash agent config parsing error")
        raise ValueError("yCrash agent config parsing error")
    return data


"""
    Validates yCrash config file 
"""
def check_config(version, k, s, a, m3Frequency):
    if not version:
        ycrash_logger.error(f"yCrash agent config Version field is empty or null")
        raise ValueError("Version field is empty or null")
    if not k:
        ycrash_logger.error(f"yCrash agent config k (key) field is empty or null")
        raise ValueError(f"yCrash agent config k (key) field is empty or null")
    if not s:
        ycrash_logger.error(f"yCrash agent config s (server url) field is empty or null")
        raise ValueError("yCrash agent config s (server url) field is empty or null")
    if not a:
        ycrash_logger.error(f"yCrash agent config a (application name) field is empty or null")
        raise ValueError(f"yCrash agent config a (application name) field is empty or null")
    if not m3Frequency:
        ycrash_logger.error(f"yCrash agent config m3Frequency field is empty or null")
        raise ValueError(f"yCrash agent config m3Frequency field is empty or null")








"""
    yCrash Profiler: Logger
    A logger to log yCash profiler activities
"""

import logging

def configure_ycrash_logger():
    y_logger = logging.getLogger('ycrash-logger')
    y_logger.setLevel('DEBUG')
    formatter = logging.Formatter('yCrash: %(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler('ycrash.log')
    file_handler.setFormatter(formatter)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    logging.getLogger().addHandler(console_handler)
    y_logger.addHandler(file_handler)
    return y_logger


ycrash_logger = configure_ycrash_logger()



"""
    yCrash System Profiler: Encapsulates all the system commands
"""

from log.ycrash_logger import ycrash_logger
import os

PING_ENDPOINT = "google.com"


"""
    Base command class containing system commands
"""
class BaseCommands:
    def __init__(self, NetStat="netstat -an", PS="ps -ef", PS2="ps -ef", M3PS="ps -ef", Disk="df -hk",
                 Top="top -l 5 && sleep 10", Top2="", TopH="top -l 1 -pid {os.getpid()}", TopH2="",
                 Top4M3="top -l 1", VMState="vm_stat -c 10 5", DMesg="dmesg",
                 DMesg2="cat /var/log/system.log | tail -20", ProcessTopCPU="ps -eo pid,command,%cpu -r",
                 ProcessTopMEM="ps -eo pid,command,%mem -m", OSVersion="uname -a", KernelParam="sysctl -a",
                 Ping=f"ping -c 6 {PING_ENDPOINT}", SHELL="/bin/sh -c"):
        self.NetStat = NetStat
        self.PS = PS
        self.PS2 = PS2
        self.M3PS = M3PS
        self.Disk = Disk
        self.Top = Top
        self.Top2 = Top2
        self.TopH = TopH
        self.TopH2 = TopH2
        self.Top4M3 = Top4M3
        self.VMState = VMState
        self.DMesg = DMesg
        self.DMesg2 = DMesg2
        self.ProcessTopCPU = ProcessTopCPU
        self.ProcessTopMEM = ProcessTopMEM
        self.OSVersion = OSVersion
        self.KernelParam = KernelParam
        self.Ping = Ping
        self.SHELL = SHELL

    """
        Get the system commands object on the os type
    """
    def get_commands(self,os_type):
        ycrash_logger.debug(f"os_type: {os_type}")
        if os_type == "Darwin":
            return DarwinCommands()
        elif os_type == "Linux":
            return LinuxCommands()
        elif os_type == "Solaris":
            return SolarisCommands()
        elif os_type == "Windows":
            return WindowsCommands()
        elif os_type == "AIX":
            return AIXCommands()
        else:
            raise NotImplementedError(f"OS type '{os_type}' not supported.")
"""
    Darwin(mac) command class containing system commands
"""
class DarwinCommands(BaseCommands):
    def __init__(self):
        super().__init__(
            NetStat="netstat -an",
            PS="ps -ef",
            PS2="ps -ef",
            M3PS="ps -ef",
            Disk="df -hk",
            Top="for i in {1..3}; do top -l 5 ; sleep 10; done",
            Top2="",  # Placeholder for no operation
            TopH=f"top -l 1 -pid {os.getpid()}",
            TopH2="",  # Placeholder for no operation
            Top4M3="top -l 1",
            VMState="vm_stat -c 10 5",
            DMesg="dmesg",
            DMesg2="cat /var/log/system.log | tail -20",
            ProcessTopCPU="ps -eo pid,command,%cpu -r",
            ProcessTopMEM="ps -eo pid,command,%mem -m",
            OSVersion="uname -a",
            KernelParam="sysctl -a",
            Ping=(f"ping -c 6 {PING_ENDPOINT}"),
            SHELL="/bin/sh -c"
        )
"""
    Linux command class containing system commands
    TODO -  Untested
"""
class LinuxCommands(BaseCommands):
    def __init__(self):
        super().__init__(
            NetStat="netstat -pan",
            PS="ps -eLf",
            PS2="ps -eTf",
            M3PS="ps -eLf",
            Disk="df -hk",
            Top="top -bc -d 10 -n 3",
            Top2="Executable() -topMode -bc -d 10 -n 3",
            TopH="top -bH -n 1 -p {DynamicArg}",
            TopH2="Executable() -topMode -bH -n 1 -p",
            Top4M3="Executable() -topMode -bc -n 1",
            #  TOCHECK - What is the TOP_INTERVAL is the frequency ? and what is script span
            # VMState="vmstat {DynamicArg} {DynamicArg} | awk '{{cmd=\"(date +'\\''%H:%M:%S'\\'' )\"; cmd | getline now; print now $0; fflush(); close(cmd)}}'",
            DMesg="dmesg -T --level=emerg,alert,crit,err,warn | tail -20",
            DMesg2="dmesg --level=emerg,alert,crit,err,warn | tail -20 | awk '{{gsub(/\\\\[[^]]*\\\\]/,\"\"); print strftime(\"[%%a %%b %%d %%H:%%M:%%S %%Y]\", systime()-$(NF-1)), $0}}'",
            ProcessTopCPU="ps -o pid,%cpu,cmd,ax | sort -b -k2 -r",
            ProcessTopMEM="ps -o pid,%mem,cmd,ax | sort -b -k2 -r",
            OSVersion="uname -a",
            KernelParam="sysctl -a",
            Ping=f"ping -c 6 {PING_ENDPOINT}",
            SHELL="/bin/sh -c",
        )

"""
    Solaris command class containing system commands
    TODO -  Untested
"""
class SolarisCommands(BaseCommands):
    def __init__(self):
        super().__init__(
            NetStat="netstat -pan",
            PS="ps -eLf",
            PS2="ps -eLf",
            M3PS="ps -eLf",
            Disk="df -hk",
            #  TOCHECK - What is the TOP_INTERVAL is the frequency ? and what is script span
            # Top=f"top -bc -d {TOP_INTERVAL} -n {SCRIPT_SPAN//TOP_INTERVAL + 1}",
            # Top2=f"top -bc -d {TOP_INTERVAL} -n {SCRIPT_SPAN//TOP_INTERVAL + 1}",
            TopH="top -bH -n 1 -p {DynamicArg}",
            TopH2="top -bH -n 1 -p {DynamicArg}",
            Top4M3="top -bc -n 1",
            #  TOCHECK - What are the dynamic arguements
            #  VMState=f"vmstat {DynamicArg} {DynamicArg} | awk '{{now=strftime(\"%T \"); print now $0; fflush()}}'",
            DMesg="dmesg",
            DMesg2="dmesg",
            ProcessTopCPU="ps -eo pid,cmd,%cpu --sort=-%cpu",
            ProcessTopMEM="ps -eo pid,cmd,%mem --sort=-%mem",
            OSVersion="uname -a",
            KernelParam="sysctl -a",
            SHELL="/bin/sh -c"
        )
"""
    AIX command class containing system commands
    TODO -  Untested
"""
class AIXCommands(BaseCommands):
    def __init__(self):
        super().__init__(
            NetStat="netstat -a",
            PS="ps -ef",
            PS2="ps -ef",
            M3PS="ps -ef",
            Disk="df",
            Top="topas -P",
            Top2="topas -P",
            TopH="topas -P",
            TopH2="topas -P",
            Top4M3="topas -P",
            #  TOCHECK - What are dynamic arguments ?
            # VMState=Command("vmstat", DynamicArg, DynamicArg),
            DMesg="dmesg",
            DMesg2="dmesg",
            ProcessTopCPU="ps -eo pid,cmd,%cpu --sort=-%cpu",
            ProcessTopMEM="ps -eo pid,cmd,%mem --sort=-%mem",
            #  TOCHECK - What is WaitCommand why is it required
            # OSVersion=Command(WaitCommand, "uname", "-a"),
            # KernelParam=Command(WaitCommand, "sysctl", "-a"),
            SHELL="/bin/sh -c"
        )

"""
    Windoes command class containing system commands
    TODO -  Untested
"""
class WindowsCommands(BaseCommands):
    def __init__(self):
        super().__init__(
            NetStat="netstat -an",
            PS="PowerShell.exe -Command Get-CimInstance -Class Win32_Process | ConvertTo-Json",
            PS2="PowerShell.exe -Command Get-CimInstance -Class Win32_Process | ConvertTo-Json",
            M3PS="PowerShell.exe -Command Get-CimInstance -Class Win32_Process | ConvertTo-Json",
            Disk="wmic logicaldisk get size,freespace,caption",
            Top="NotSupported",
            Top2="NotSupported",  # Placeholder for no operation
            TopH="NotSupported",
            TopH2="NotSupported",  # Placeholder for no operation
            Top4M3="NotSupported",
            VMState= "PowerShell.exe -Command & {typeperf -sc 10 -si 5 \System\Processor Queue Length \PhysicalDisk(_Total)\Current Disk Queue Length \Process(_Total) \Page File Bytes \Memor \Available KBytes \Memory \Modified Page List Bytes \Memory \Cache Bytes \Memory \Pages Input/sec \Memory \Pages Output/sec \PhysicalDisk(_Total) \Disk Transfers/sec \PhysicalDisk(_Total) \Disk Writes/sec \Processor(_Total) \Interrupts/sec \System \Context Switches/sec \Processor(_Total) \% User Time \Processor(_Total) \% Privileged Time \Processor(_Total) \% Idle Time \Processor(_Total) \% Interrupt Time \Processor(_Total) \% DPC Time'}",
            DMesg="PowerShell.exe -Command & {Get-EventLog -LogName System -Newest 20 -EntryType Error,FailureAudit,Warning | Select-Object TimeGenerated, EntryType, Message | ForEach-Object { Write-Host $($_.TimeGenerated) [$($_.EntryType)]: $($_.Message) }}",
            DMesg2="wevtutil  qe System /c:20 /rd:true /f:text",
            ProcessTopCPU="PowerShell.exe -Command & {ps | sort -desc CPU}",
            ProcessTopMEM="PowerShell.exe -Command & {ps | sort -desc PM}",
            OSVersion="PowerShell.exe -Command & {systeminfo | findstr /B /C:OS Name /C:OS Version}",
            KernelParam="NotSupported",
            Ping=f"ping -c 6 {PING_ENDPOINT}",
            SHELL="cmd.exe /c"
                )





import requests
from datetime import datetime
import socket
from log.ycrash_logger import ycrash_logger

"""
    yCrash Profiler: WebClient
    Uploads json request to the server url.
"""
def upload_json_data(contentType, apiKey, url, data):

    # Get the current time
    current_time = datetime.now()

    # Format the time as required
    formatted_time = current_time.strftime("%Y-%m-%dT%H:%M:%S")
    hostname = socket.gethostname()
    ip_address = socket.gethostbyname(hostname)

    url = url+"?de="+ip_address+"?ts="+formatted_time+"&lang=py"
    # Set the Content-Type header to application/json
    headers = {'Content-Type': contentType,
               'Authorization': apiKey}
    ycrash_logger.debug(f"\n*****************************************************************************************")
    ycrash_logger.debug(f" \nHTTP post to  {url}  with headers {headers}")
    ycrash_logger.debug(f" DATA: ")
    ycrash_logger.debug(f" {data}")
    response = requests.post(url, data=data, headers=headers)
    ycrash_logger.debug(f"Response: {response}")
    if response.status_code == 200:
        ycrash_logger.debug('Data uploaded successfully!')
    else:
        ycrash_logger.debug('Error uploading data:', response.status_code)
    ycrash_logger.debug(f"*****************************************************************************************")
    # Send the POST request with JSON data

    # Check if the request was successful




import sys
import traceback
import threading
import time
import json
import psutil
from commons.webclient import upload_json_data
from commons.common import  getKey, getServerUrl

THREAD_DUMP_ENDPOINT = "/cm-receiver+&dt=td"

def capture_thread_states():
    threads = {}
    # Capture states and stack traces of all threads
    for thread_id, frame in sys._current_frames().items():
        thread = get_thread_by_id(thread_id)
        # print_thread_attributes(thread)
        event = get_event(thread)
        proc = psutil.Process()

        # print(f"event {event}")
        threads[thread_id] = {
            'name': thread.name,
            'native_id': thread.native_id,
            'details': vars_to_string(thread),
            'cpu_percent':  get_threads_cpu_percent(proc,thread_id),
            'event': event,
            'state': get_thread_state(thread),
            'is_waiting': is_waiting_for_lock(frame),
            'stack_trace': traceback.format_stack(frame)
        }
    return threads


def vars_to_string(obj):
    values_str = ""
    for value in vars(obj).values():
        values_str += str(value) + " "
    return values_str


def get_event(thread):
    event_value = ""
    for obj in vars(thread).values():
        if type(obj) is threading.Event:
            print(event_value)
            event_value = event_value + vars_to_string(obj)
    return event_value


def print_thread_attributes(thread):
    thread_vars = vars(thread)
    for attr_name, attr_value in thread_vars.items():
        print(f"Attribute: {attr_name}, Type: {type(attr_value)}, Value: {attr_value}")


def is_waiting_for_lock(frame):
    stack_trace = traceback.format_stack(frame)
    for line in stack_trace:
        if '_wait_for_tstate_lock' in line:
            return True
    return False


def find_locks_held_by_thread(thread_id):
    locks = set()
    for obj in threading._active.values():
        if type(obj) is threading.Lock:
            # If the lock is owned by the specified thread, add it to the set of locks
            if obj._is_owned(thread_id):
                locks.add(obj)

    return locks


def get_thread_state(thread):
    if thread.is_alive():
        return "Alive"
    if thread.daemon:
        return "Dead (Daemon)"
    if thread.ident is None:
        return "New"
    return "UNKNOWN"


def get_thread_by_id(thread_id):
    for thread in threading.enumerate():
        if thread.ident == thread_id:
            return thread
    return None


def capture_export_thread_data(config):
    threads = capture_thread_states()
    condition = threading.Condition()
    current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    thread_events = []
    for thread_id, thread_info in threads.items():
        thread_data = {
            "event_type": "THREAD_DETAILS",
            "native_id": thread_info['native_id'],
            "id": thread_id,
            "name": thread_info['name'],
            "state": thread_info['state'],
            # "cpu_percent": thread_info['cpu_percent'],
            "waiting": thread_info['is_waiting'],
            # "details": thread_info['details'],
            "thread_event": thread_info['event'],
            "stack_trace": [line.strip() for line in thread_info['stack_trace']]
        }
        thread_events.append(thread_data)
        thread = {
                "timeStamp": current_time,
                "threads": [thread_events],  # Dictionary with threadevents list as a value
        }
        thread_json = json.dumps(thread, indent=4)
    upload_json_data('application/json', getKey(config), getServerUrl(config,THREAD_DUMP_ENDPOINT), thread_json)


def get_locks_and_threads():
    lock_threads = {}
    for thread_id, thread in threading._active.items():
        for lock in threading._active[thread_id]._Thread__daemonic_locks:
            lock_id = id(lock)
            if lock_id not in lock_threads:
                lock_threads[lock_id] = []
            lock_threads[lock_id].append(thread)

    return lock_threads


def is_thread_blocked_or_waiting(thread):
    # Check if the thread is blocked on a lock
    for lock in threading._active_limbo_locks:
        if thread in lock._waiters:
            return True

    # Check if the thread is blocked on an event
    for event in threading._active_limbo_events:
        if thread in event._cond._waiters:
            return True

    # Add more checks for other synchronization primitives as needed...

    return False


def find_locks_held_by_threadold(thread_id):
    locks = []

    # Iterate through all objects in threading._active, which includes all thread objects
    for obj in threading._active.values():
        # Check if the object has a _is_owned() method (indicating it's a lock)
        if hasattr(obj, '_is_owned') and callable(obj._is_owned):
            # If the lock is owned by the specified thread, add it to the list of locks
            if obj._is_owned(thread_id):
                locks.append(obj)

    return locks


def get_threads_cpu_percent(p, thread_id, interval=0.1):
    total_percent = p.cpu_percent(interval)
    total_time = sum(p.cpu_times())
    cpu_percentage = 0
    for t in p.threads():
          if t.id == thread_id:
            cpu_percentage = total_percent * ((t.system_time + t.user_time) / total_time)
    return cpu_percentage



"""
    yCrash System Profiler: System information
"""
import platform

class SystemInfo:
    def __init__(self):
        self.os_name = platform.system()
        self.architecture = platform.architecture()[0]

    def get_os_name(self):
        return self.os_name

    def get_architecture(self):
        return self.architecture





from setuptools import setup, find_packages

setup(
    name='ycrash-profiler-dev',
    version='0.7',
    packages=find_packages(exclude=['tests*']),
    description='yCrash Python Agent',
    install_requires=['pyyaml', 'requests', 'memory_profiler','psutil'],
    url='https://ycrash.io',
    author='YCrash Team',
    author_email='team@tier1app.com'
)



"""
    yCrash System Profiler: System Profiler
    This service is reponsible for retrieving and uploading system data to yCrash
"""
from datetime import datetime

from commons.common import run_command
from commons.webclient import upload_json_data
import json
from commons.common import getKey, getServerUrl
from .system_info import SystemInfo
from .system_commands import BaseCommands
from log.ycrash_logger import ycrash_logger


PS_DUMP_ENDPOINT = "/cm-receiver&dt=ps"
DF_DUMP_ENDPOINT = "/cm-receiver&dt=df"
NETSTAT_DUMP_ENDPOINT = "/cm-receiver&dt=ns"
KERNAL_DUMP_ENDPOINT = "/cm-receiver&dt=kernal"
TOP_DUMP_ENDPOINT = "/cm-receiver&dt=top"
TOPH_DUMP_ENDPOINT = "/cm-receiver&dt=toph"
DMSEG_DUMP_ENDPOINT = "/cm-receiver&dt=dmesg"
PING_DUMP_ENDPOINT = "/cm-receiver&dt=ping"
VMSTAT_DUMP_ENDPOINT = "/cm-receiver&dt=vmstat"


"""
    Runs all system commands for the OS.
"""
def export_all_system_data(config):
    system_info = SystemInfo()
    commands_manager = BaseCommands().get_commands(system_info.get_os_name())
    capture_and_upload_system_data(config, PS_DUMP_ENDPOINT, commands_manager.PS)
    capture_and_upload_system_data(config, DF_DUMP_ENDPOINT, commands_manager.Disk)
    capture_and_upload_system_data(config, NETSTAT_DUMP_ENDPOINT, commands_manager.NetStat)
    capture_and_upload_system_data(config, TOP_DUMP_ENDPOINT, commands_manager.Top)
    capture_and_upload_system_data(config, TOPH_DUMP_ENDPOINT, commands_manager.TopH)
    capture_and_upload_system_data(config, DMSEG_DUMP_ENDPOINT, commands_manager.DMesg2)
    capture_and_upload_system_data(config, KERNAL_DUMP_ENDPOINT, commands_manager.KernelParam)
    capture_and_upload_system_data(config, VMSTAT_DUMP_ENDPOINT, commands_manager.VMState)
    capture_and_upload_system_data(config, PING_DUMP_ENDPOINT, commands_manager.Ping)


"""
    Executes each system data and export the data to yCrash server
"""
def capture_and_upload_system_data(config, endpoint, command):
    start_time = datetime.now()
    data = {
        "data": run_command(command),
    }
    end_time = datetime.now()
    time_took = end_time - start_time
    system_details_json = json.dumps(data, indent=4)
    upload_json_data('application/json', getKey(config), getServerUrl(config, endpoint),
                     system_details_json)
    ycrash_logger.debug(f"Time taken for command {command} is {time_took.total_seconds()}")


"""
    yCrash Python Profiler: Heap Dump Analyser
    This service responsible to extract and heap dumps upload to yCrash Server
    Dependency : Requries guppy module
"""

import tracemalloc
import json
from guppy import hpy
from commons.webclient import upload_json_data
from commons.common import  getKey, getServerUrl
from gcprofiler.gc_profiler import ycrash_profile

HEAP_DUMP_ENDPOINT = "/cm-receiver+&dt=hd"

@ycrash_profile
class HeapDumpCapturer:
    def __init__(self):
        tracemalloc.start()

    """
    Captures heap dump in two formats and uploads to server
    """

    def capture_heap_dump(self, config):
       #print(f"YC Config:{config}")
       heapDumpJson = {'heapDumpFormat1':
                        capture_heap_tracemalloc_dump(self),
                       'heapDumpFormat2':
                        f"\'{capture_heap_guppy_dump(self)}\'"
                        }
       upload_json_data('application/json', getKey(config), getServerUrl(config,HEAP_DUMP_ENDPOINT), heapDumpJson)
       #print(f'uploaded json {heapDumpJson} with key {key}')

"""
    Captures heap tracemalloc dumps, providing a configured number of objects along with their references and sizes.
"""
def capture_heap_tracemalloc_dump(self, top_n=10):
        snapshot = tracemalloc.take_snapshot()
        stats = snapshot.statistics('lineno')

        # Convert stats to a list of dictionaries
        heap_dump = [{
            'filename': stat.traceback[0].filename,
            'lineno': stat.traceback[0].lineno,
            'sizeInBytes': stat.size,
            'noOfTimesCalled': stat.count
        } for stat in stats[:top_n]]

        heapDumpList = []
        for entry in heap_dump:
            heapDumpList.append(json.dumps(entry, indent=4))
        return heapDumpList

"""
    Captures heap dump with guppy with providing objects and sizes
"""
def capture_heap_guppy_dump(self):
    hp = hpy()
    heap: object = hp.heap()
    print(heap)
    return heap



"""
    yCrash Python Profiler:GC/Mem Profiler
    This service responsible extracting and uplading the GC and memory data to yCrash server
"""

from memory_profiler import memory_usage
import types
import json
from commons.webclient import upload_json_data
from commons.common import getKey, getServerUrl
import gc
import inspect
import re
from collections import Counter
import sys

GC_DUMP_ENDPOINT = "/cm-receiver+&dt=gc"


"""
   The func retrieved and uploads GC stats, details and leaked objects to the yCrash server
"""
def ycrash_memory_extract(config):
    mem_usage = memory_usage()[0]
    gc_counts = getCounts(gc.get_count())
    gc_count_data = {
        "young": gc_counts[0],
        "middle": gc_counts[1],
        "old": gc_counts[2]
    }

    gc_threshold = getCounts(gc.get_threshold())
    gc_threshold_data = {
        "young": gc_threshold[0],
        "middle": gc_threshold[1],
        "old": gc_threshold[2]
    }

    stats = gc.get_stats()
    gc_stats_data = {
        "young": stats[0],
        "middle": stats[1],
        "old": stats[2]
    }

    stats = gc.get_stats()
    topObjects = getTopObjects()
    gc_stats = {"gc_top_objects": topObjects}

    leaked_objects = []

    # Check for objects in gc.garbage
    if gc.garbage:
        for obj in gc.garbage:
            leaked_objects.append(repr(obj))

    freeze_count = gc.get_freeze_count()

    print(stats)
    data = {"memory_usage": mem_usage,
            "gc": {"gc_counts": gc_count_data, "gc_threshold": gc_threshold_data, "gc_stats": gc_stats_data,
                   'gc_top_objects': topObjects, "leaked_objects": leaked_objects, "freeze_count": freeze_count}}

    # Get a list of all objects tracked by the garbage collector
    # Create a dictionary to store the leaked objects

    gc.get_freeze_count()

    memory_json = json.dumps(data, indent=4)
    upload_json_data('application/json', getKey(config),
                     getServerUrl(config, GC_DUMP_ENDPOINT), memory_json)

"""
   Retrieves and returns the top objects 
"""

def getTopObjects():
    objects = gc.get_objects()

    # Extract class names of objects and count occurrences
    class_names = [obj.__class__.__name__ for obj in objects]
    class_counts = Counter(class_names)

    # Sort the Counter by counts in descending order and limit to top 10
    top_10_classes = class_counts.most_common(30)

    # Create a list of dictionaries
    result_list = [{"class_name": name, "count": count} for name, count in top_10_classes]

    # Convert the list to JSON
    json_data = json.dumps(result_list)
    return json_data


def getCounts(gc_counts):
    counts_str = str(gc_counts)
    # Use regular expression to extract numbers
    counts = tuple(map(int, re.findall(r'\d+', counts_str)))
    return counts


"""
    Decorates a function to captures  memory stats
"""
def ycrash_profile_memory(func):
    def wrapper(*args, **kwargs):
        mem_usage_start = memory_usage()[0]
        result = func(*args, **kwargs)
        mem_usage_end = memory_usage()[0]
        mem_usage_diff = mem_usage_end - mem_usage_start

        print(f"Memory usage for {func.__name__}: {mem_usage_diff} MB")
        return result

    return wrapper


"""
    Finds all methods in the current modules
"""
def find_all_modules():
    for module_name, module_obj in sys.modules.items():
        print(f"*** module is {module_name} {module_obj}")
        profile_global_functions(module_obj)
        profile_classes(module_obj)


def profile_global_functions(module):
    for name, func in module.__dict__.items():
        if callable(func) and not name.startswith('__') and not name.startswith('ycrash_profile'):
            setattr(module, name, ycrash_profile_memory(func))
    return module

"""
    Decorates all class in a module for capturing memory usage
"""
def profile_classes(module):
    for name, obj in vars(module).items():
        if inspect.isclass(obj):
            # Decorate each method in the class
            for method_name, method in vars(obj).items():
                if callable(method):
                    setattr(obj, method_name, ycrash_profile_memory(method))

"""
    Decorates all methods in class in a module for capturing memory usage
"""
# Annotate the functions with their types
def ycrash_profile(cls):
    for name, method in vars(cls).items():
        if isinstance(method, types.FunctionType):
            setattr(cls, name, ycrash_profile_memory(method))
    return cls

"""
    Decorates all methods in class in a current application for capturing memory usage
"""
def profile_all_methods():
    # Get all functions defined in the current module
    current_module = __import__(__name__)
    functions = [obj for name, obj in current_module.__dict__.items() if isinstance(obj, types.FunctionType)]

    # Apply ycrash_profile_memory decorator to each function
    for func in functions:
        setattr(current_module, func.__name__, ycrash_profile_memory(func))



"""
    yCrash Python Profiler: Process Analyser
    This service responsible to extract the process details and uploads the data to sever
"""

import psutil
import json
from commons.common import get_current_pid
from commons.webclient import upload_json_data
from commons.common import  getKey, getServerUrl
import getpass
import logger
import socket
from datetime import datetime
import time
import os
import sys

PROCESS_DUMP_ENDPOINT = "/cm-receiver+&dt=meta"

"""
    Retrieves and uploads the process details to the yCrash server
"""
def capture_print_process_details(config):
    try:
        process = psutil.Process(get_current_pid())
        process_data = {
            "pid": get_current_pid(),
            "host_name" : socket.gethostname(),
            "who_am_i": getpass.getuser(),
            "cpu_count": os.cpu_count(),
            "python_version": sys.version,
            "cpu_percent": process.cpu_percent(interval=1),
            "memory_info": process.memory_info()._asdict(),
            "timestamp": datetime.now().strftime("%Y-%m-%dT%H-%M-%S"),
            "timezone": get_system_timezone(),
            "connections": len(process.connections()),
            "threads_count": len(process.threads()) # Count the number of threads
        }
        process_details = {"process_details": process_data}
        json_data = json.dumps(process_details, indent=4)
        upload_json_data('application/json', getKey(config), getServerUrl(config,PROCESS_DUMP_ENDPOINT), json_data)
    except psutil.NoSuchProcess:
        logger.error(f"Process with PID {get_current_pid()} not found.")
    except Exception as e:
        logger.error(f"Error: {e}")

"""
    Retrieves the IO details
"""
def get_io_details(io_counter):
    try:
        # Create a dictionary with I/O details
        io_dict = {
            "read_count": io_counter.read_count,
            "write_count": io_counter.write_count,
            "read_bytes": io_counter.read_bytes,
            "write_bytes": io_counter.write_bytes
        }
        # Convert the dictionary to a JSON-formatted string
        io_json = json.dumps(io_dict, indent=4)
        return io_json
    except psutil.Error as e:
        # Handle potential exceptions from psutil
        print(f"Error getting I/O details: {e}")
        return None

"""
    Returns the thread CPU usage percentage
"""
def get_threads_cpu_percent(p):
    threadList = []
    for thread in p.threads():
        total_percent = p.cpu_percent(0.1)
        total_time = sum(p.cpu_times())
        cpu_percentage = total_percent * ((thread.system_time + thread.user_time) / total_time)
        process_data = {'id': thread.id,
                        'cpu_percentage': cpu_percentage}

        threadList.append(process_data)
    return threadList

"""
    Retrieves the system timezone
"""
def get_system_timezone():
    # Get the timezone offset in seconds
    timezone_offset_seconds = time.timezone

    # Convert the offset to hours and minutes
    hours = abs(timezone_offset_seconds) // 3600
    minutes = (abs(timezone_offset_seconds) % 3600) // 60

    # Determine the sign of the timezone offset
    sign = '+' if timezone_offset_seconds >= 0 else '-'

    # Format the timezone information
    timezone_info = f"UTC{sign}{hours:02d}:{minutes:02d}"

    return timezone_info



import threading
import logging


def sampleLogging():
    buggy_app_logger.debug('This is log simulation')
    buggy_app_logger.info('This is log simulation')
    buggy_app_logger.warning('This is log simulation')
    buggy_app_logger.error('This is log simulation')
    buggy_app_logger.critical('This is log simulation')


def contineous_logs_simulation():
    logging_thread = threading.Thread(target=sampleLogging, name="buggyapp.logger.thread")
    logging_thread.start()


def configure_buggyapp_logger():
    b_logger = logging.getLogger('buggyapp-logger')
    b_logger.setLevel('INFO')
    formatter = logging.Formatter('buggyapp: %(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler = logging.FileHandler('buggyapp.log')
    file_handler.setFormatter(formatter)
    b_logger.addHandler(file_handler)
    return b_logger


buggy_app_logger = configure_buggyapp_logger()
